# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import logging
from os.path import join, dirname, abspath, isdir, isfile

lib_logger = logging.getLogger('pypdfium2')
lib_logger.addHandler(logging.StreamHandler())

TestDir     = dirname(abspath(__file__))
SourceTree  = dirname(TestDir)
ResourceDir = join(TestDir,'resources')
OutputDir   = join(TestDir,'output')

sys.path.insert(0, join(SourceTree,'setupsrc'))
from pl_setup.packaging_base import PlatformNames


class TestFiles:
    render        = join(ResourceDir,'render.pdf')
    encrypted     = join(ResourceDir,'encrypted.pdf')
    multipage     = join(ResourceDir,'multipage.pdf')
    toc           = join(ResourceDir,'toc.pdf')
    toc_viewmodes = join(ResourceDir,'toc_viewmodes.pdf')
    toc_maxdepth  = join(ResourceDir,'toc_maxdepth.pdf')
    toc_circular  = join(ResourceDir,'toc_circular.pdf')
    box_fallback  = join(ResourceDir,'box_fallback.pdf')
    text          = join(ResourceDir,'text.pdf')
    empty         = join(ResourceDir,'empty.pdf')
    images        = join(ResourceDir,'images.pdf')


ExpRenderPixels = (
    ( (0,   0  ), (255, 255, 255) ),
    ( (150, 180), (129, 212, 26 ) ),
    ( (150, 390), (42,  96,  153) ),
    ( (150, 570), (128, 0,   128) ),
)


def iterate_testfiles(skip_encrypted=True):
    encrypted = (TestFiles.encrypted, )
    for attr_name in dir(TestFiles):
        if attr_name.startswith('_'):
            continue
        member = getattr(TestFiles, attr_name)
        if skip_encrypted and member in encrypted:
            continue
        yield member


def _get_attrs(cls):
    members = []
    for attr in dir(cls):
        if attr.startswith('_'):
            continue
        members.append(attr)
    return members

def get_members(cls):
    return [getattr(cls, a) for a in _get_attrs(cls)]

pl_names = _get_attrs(PlatformNames)


def test_testpaths():
    for dirpath in (TestDir, SourceTree, ResourceDir, OutputDir):
        assert isdir(dirpath)
    for filepath in iterate_testfiles(False):
        assert isfile(filepath)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import ast
from os.path import (
    join,
    abspath,
    basename,
    splitext,
)
from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parsers import pagetext_type


def rotation_type(string):
    rotation = int(string)
    if rotation not in (0, 90, 180, 270):
        raise ValueError("Invalid rotation value %s" % rotation)
    return rotation


def colour_type(string):
    if string.lower() == "none":
        return
    else:
        colour = ast.literal_eval(string)
        if not isinstance(colour, (tuple, list)):
            raise ValueError("Invalid colour type %s. Must be list or tuple." % type(colour))
        if not len(colour) in (3, 4):
            raise ValueError("Invalid number of colour values. Must be 3 or 4.")
        if not all(isinstance(val, int) and 0 <= val <= 255 for val in colour):
            raise ValueError("Colour values must be integers ranging from 0 to 255.")
        return colour


def crop_type(string):
    crop = ast.literal_eval(string)
    if not isinstance(crop, (tuple, list)) or len(crop) != 4 or not all(isinstance(c, (int, float)) for c in crop):
        raise ValueError("Crop must be a list of four numbers.")
    return crop


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "render",
        help = "Rasterise pages of a PDF file",
    )
    parser.add_argument(
        "inputs",
        nargs = "+",
        help = "PDF documents to render",
    )
    parser.add_argument(
        "--output", "-o",
        type = abspath,
        required = True,
        help = "Output directory where to place the serially numbered images",
    )
    parser.add_argument(
        "--format", "-f",
        default = "jpg",
        help = "File extension of the image format to use",
    )
    parser.add_argument(
        "--pages",
        default = None,
        type = pagetext_type,
        help = "Numbers of the pages to render (defaults to all)",
    )
    parser.add_argument(
        "--scale",
        default = 1,
        type = float,
        help = "Define the resolution of the output images. By default, one PDF point (1/72in) is rendered to 1x1 pixel. This factor scales the number of pixels that represent one point.",
    )
    parser.add_argument(
        "--rotation",
        default = 0,
        type = rotation_type,
        help = "Rotate pages by 90, 180 or 270 degrees",
    )
    parser.add_argument(
        "--colour",
        default = (255, 255, 255, 255),
        type = colour_type,
        help = "Page background colour. Defaults to white. It can be given in RGBA format as a sequence of integers ranging from 0 to 255, or it may be 'none' for transparent background."
    )
    parser.add_argument(
        "--no-annotations",
        action = "store_true",
        help = "Option to prevent rendering of PDF annotations",
    )
    parser.add_argument(
        "--optimise-mode",
        default = pdfium.OptimiseMode.NONE,
        type = lambda string: pdfium.OptimiseMode[string.upper()],
        help = "Select a rendering optimisation mode (none, lcd_display, printing)",
    )
    parser.add_argument(
        "--greyscale",
        action = "store_true",
        help = "Whether to render in greyscale mode (no colours)",
    )
    parser.add_argument(
        "--crop",
        type = crop_type,
        default = (0, 0, 0, 0),
        help = "Amount to crop from (left, bottom, right, top)"
    )
    parser.add_argument(
        "--no-antialias",
        nargs = "+",
        default = (),
        choices = ("text", "image", "path"),
        help = "Item types that shall not be smoothed",
    )
    parser.add_argument(
        "--processes",
        default = os.cpu_count(),
        type = int,
        help = "The number of processes to use for rendering (defaults to the number of CPU cores)"
    )


def main(args):
        
    for input_path in args.inputs:
        
        pdf = pdfium.PdfDocument(input_path)
        if args.pages:
            page_indices = args.pages
        else:
            page_indices = [i for i in range(len(pdf))]
        
        n_digits = len(str( max(page_indices)+1 ))
        
        renderer = pdf.render_topil(
            page_indices = page_indices,
            scale = args.scale,
            rotation = args.rotation,
            crop = args.crop,
            colour = args.colour,
            annotations = not args.no_annotations,
            greyscale = args.greyscale,
            optimise_mode = args.optimise_mode,
            no_antialias = args.no_antialias,
            n_processes = args.processes,
        )
        
        prefix = splitext(basename(input_path))[0] + "_"
        for image, index in zip(renderer, page_indices):
            suffix = str(index+1).zfill(n_digits)
            output_path = "%s.%s" % (join(args.output, prefix+suffix), args.format)
            image.save(output_path)
            image.close()
        
        pdf.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import enum
import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers._utils import ErrorMapping


class FileAccess (enum.Enum):
    """
    Different ways how files can be loaded.
    
    .. list-table:: Overview of file access modes
        :header-rows: 1
        :widths: auto
        
        * - Mode
          - PDFium loader
          - Comment
        * - :attr:`.NATIVE`
          - :func:`.FPDF_LoadDocument`
          - File access managed by PDFium in C/C++.
        * - :attr:`.BUFFER`
          - :func:`.FPDF_LoadCustomDocument`
          - Data read incrementally from Python file buffer.
        * - :attr:`.BYTES`
          - :func:`.FPDF_LoadMemDocument`
          - Data loaded into memory and passed to PDFium at once.
    """
    NATIVE = 0
    BUFFER = 1
    BYTES  = 2


class OptimiseMode (enum.Enum):
    """ Modes defining how page rendering shall be optimised. """
    NONE = 0         #: No optimisation.
    LCD_DISPLAY = 1  #: Optimise for LCD displays (via subpixel rendering).
    PRINTING = 2     #: Optimise for printing.


class OutlineItem:
    """
    Class to store information about an entry in the table of contents ("bookmark").
    
    Parameters:
        level (int):
            Number of parent items.
        title (str):
            String of the bookmark.
        is_closed (bool):
            If :data:`True`, child items shall be hidden by default.
        page_index (int):
            Zero-based index of the page the bookmark points to.
        view_mode (int):
            A view mode constant (``PDFDEST_VIEW_...``) defining how the coordinates of *view_pos* shall be interpreted.
        view_pos (typing.Sequence[float]):
            Target position on the page the viewport should jump to when the bookmark is clicked.
            It is a sequence of :class:`float` values in PDF canvas units.
            Depending on *view_mode*, it can contain between 0 and 4 coordinates.
    """
    
    def __init__(
            self,
            level,
            title,
            is_closed,
            page_index,
            view_mode,
            view_pos,
        ):
        self.level = level
        self.title = title
        self.is_closed = is_closed
        self.page_index = page_index
        self.view_mode = view_mode
        self.view_pos = view_pos


class PdfiumError (RuntimeError):
    """ A PDFium exception forwarded by :func:`.raise_error` """
    pass


def raise_error(msg):
    """
    Raise a :class:`.PdfiumError` annotated with the description of PDFium's current error code.
    
    Warning:
        This shall be called only if a PDFium call failed according to return code **and**
        :func:`.FPDF_GetLastError()` is mentioned in the documentation for the function in question.
        Currently, :func:`.raise_error` is only applicable to document loading APIs.
    
    Parameters:
        msg (str): The error message explaining the problem.
    """
    
    err_code = pdfium.FPDF_GetLastError()
    pdfium_msg = ErrorMapping.get(err_code, "Error code %s" % err_code)
    
    raise PdfiumError("%s (PDFium: %s)" % (msg, pdfium_msg))



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
from ctypes import c_double
import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers.misc import PdfiumError


class PdfTextPage:
    """ Text extraction helper class. """
    
    def __init__(self, textpage, page):
        self._textpage = textpage
        self._page = page
    
    @property
    def raw(self):
        """ FPDF_TEXTPAGE: The raw PDFium text page object handle. """
        return self._textpage
    
    @property
    def page(self):
        """ PdfPage: The page this text page belongs to. """
        return self._page
    
    def close(self):
        """
        Close the text page to release allocated memory.
        This method shall be called when finished working with the text page.
        """
        pdfium.FPDFText_ClosePage(self._textpage)
    
    
    def get_text(self, left=0, bottom=0, right=0, top=0):
        """
        Extract text from given boundaries. If *right* and/or *top* are 0, they default to page width or height, respectively.
        
        See `this benchmark <https://github.com/py-pdf/benchmarks>`_ for a performance and quality comparison with other tools.
        
        Returns:
            str: The text on the page area in question, or an empty string if no text was found.
        """
        
        width, height = self.page.get_size()
        if right == 0:
            right = width
        if top == 0:
            top = height
        
        if not (0 <= left < right <= width) or not (0 <= bottom < top <= height):
            raise ValueError("Invalid page area requested.")
        
        args = (self._textpage, left, top, right, bottom)
        n_chars = pdfium.FPDFText_GetBoundedText(*args, None, 0)
        if n_chars <= 0:
            return ""
        
        c_array = (ctypes.c_ushort * (n_chars+1))()
        pdfium.FPDFText_GetBoundedText(*args, ctypes.cast(c_array, ctypes.POINTER(ctypes.c_ushort)), n_chars)
        text = bytes(c_array).decode("utf-16-le", errors="ignore")[:-1]
        
        return text
    
    
    def count_chars(self):
        """
        Returns:
            int: The number of characters on the page.
        """
        return pdfium.FPDFText_CountChars(self._textpage)
    
    
    def count_rects(self, index=0, count=0):
        """
        Parameters:
            index (int): Character index at which to start.
            count (int): Character count to consider (defaults to :meth:`.count_chars`).
        Returns:
            int: The number of text rectangles on the page.
        """
        n_chars = self.count_chars()
        if n_chars == 0:
            return 0
        if count == 0:
            count = n_chars
        if not (0 <= index < index+count <= n_chars):
            raise ValueError("Character span is out of bounds.")
        return pdfium.FPDFText_CountRects(self._textpage, index, count)
    
    
    def get_index(self, x, y, x_tol, y_tol):
        """
        Get the character index for a given position. All coordinates and lengths are to be given in PDF canvas units.
        
        Parameters:
            x (float): Horizontal position.
            y (float): Vertical position.
            x_tol (float): Horizontal tolerance.
            y_tol (float): Vertical tolerance.
        
        Returns:
            typing.Optional[int]: The index of the character at or nearby the point (x, y).
            May be :data:`None` if there is no character or an error occurred.
        """
        index = pdfium.FPDFText_GetCharIndexAtPos(self._textpage, x, y, x_tol, y_tol)
        if index < 0:
            return None
        return index
    
    
    def get_charbox(self, index, loose=False):
        """
        Get the bounding box of a single character.
        
        Parameters:
            index (int):
                Slot of the character to work with, in the page's character array.
            loose (bool):
                If True, the entire glyph bounds will be covered, without taking the actual glyph shape into account.
        Returns:
            (float, float, float, float):
            Values for left, bottom, right and top in PDF canvas units.
        """
        
        n_chars = self.count_chars()
        if not 0 <= index < n_chars:
            raise ValueError("Character index %s is out of bounds. The maximum index is %d." % (index, n_chars-1))
        
        if loose:
            rect = pdfium.FS_RECTF()
            ret_code = pdfium.FPDFText_GetLooseCharBox(self._textpage, index, rect)
            left, bottom, right, top = rect.left, rect.bottom, rect.right, rect.top
        else:
            left, bottom, right, top = c_double(), c_double(), c_double(), c_double()
            ret_code = pdfium.FPDFText_GetCharBox(self._textpage, index, left, right, bottom, top)
            left, bottom, right, top = left.value, bottom.value, right.value, top.value
        
        if not ret_code:
            raise PdfiumError("Retrieving the char box failed")
        
        return left, bottom, right, top
    
    
    def get_rectboxes(self, index=0, count=0):
        """
        Get the bounding boxes of text rectangles in the requested scope.
        
        Yields:
            Coordinates for left, bottom, right, and top (as :class:`float` values).
        """
        n_rects = self.count_rects(index, count)
        for index in range(n_rects):
            left, top, right, bottom = c_double(), c_double(), c_double(), c_double()
            pdfium.FPDFText_GetRect(self._textpage, index, *[ctypes.byref(coord) for coord in (left, top, right, bottom)])
            yield (left.value, bottom.value, right.value, top.value)
    
    
    def get_links(self):
        """
        Iterate through web links on the page.
        
        Yields:
            :class:`str`: A web link string.
        """
        
        links = pdfium.FPDFLink_LoadWebLinks(self._textpage)
        n_links = pdfium.FPDFLink_CountWebLinks(links)
        
        for i in range(n_links):
            n_chars = pdfium.FPDFLink_GetURL(links, i, None, 0)
            buffer = (ctypes.c_ushort * n_chars)()
            pdfium.FPDFLink_GetURL(links, i, buffer, n_chars)
            link = bytes(buffer).decode("utf-16-le")[:-1]
            yield link
        
        pdfium.FPDFLink_CloseWebLinks(links)
    
    
    def search(self, text, index=0, match_case=False, match_whole_word=False):
        """
        Locate text on the page.
        
        Parameters:
            text (str):
                The string to search for.
            index (int):
                Character index at which to start searching.
            match_case (bool):
                If :data:`True`, the search will be case-specific (upper and lower letters treated as different characters).
            match_whole_word (bool):
                If :data:`True`, substring occurrences will be ignored (e. g. `cat` would not match `category`).
        Returns:
            PdfTextSearcher: A helper object to search text.
        """
        
        if len(text) == 0:
            raise ValueError("Text length must be >0.")
        
        flags = 0
        if match_case:
            flags |= pdfium.FPDF_MATCHCASE
        if match_whole_word:
            flags |= pdfium.FPDF_MATCHWHOLEWORD
        
        enc_text = text.encode("utf-16-le") + b"\x00\x00"
        text_pointer = ctypes.cast(enc_text, ctypes.POINTER(ctypes.c_uint16))
        search = pdfium.FPDFText_FindStart(self._textpage, text_pointer, flags, index)
        return PdfTextSearcher(search, self)
        

class PdfTextSearcher:
    """ Text searcher helper class. """
    
    def __init__(self, search, textpage):
        self._search = search
        self._textpage = textpage
    
    @property
    def raw(self):
        """ FPDF_SCHHANDLE: The raw PDFium searcher handle. """
        return self._search
    
    @property
    def textpage(self):
        """ PdfTextPage: The text page this searching helper belongs to. """
        return self._textpage
    
    def _get_occurrence(self, find_func):
        found = find_func(self._search)
        if not found:
            return None
        index = pdfium.FPDFText_GetSchResultIndex(self._search)
        count = pdfium.FPDFText_GetSchCount(self._search)
        return tuple( [box for box in self._textpage.get_rectboxes(index, count)] )
    
    def get_next(self):
        """
        Returns:
            typing.Sequence[ typing.Tuple[float, float, float, float] ]:
            A list of bounding boxes for the next occurrence, or :data:`None` if the last occurrence was passed.
        """
        return self._get_occurrence(pdfium.FPDFText_FindNext)
    
    def get_prev(self):
        """
        Returns:
            typing.Sequence[ typing.Tuple[float, float, float, float] ]:
            A list of bounding boxes for the previous occurrence (i. e. the one before the last valid occurrence),
            or :data:`None` if the first occurrence was passed.
        """
        return self._get_occurrence(pdfium.FPDFText_FindPrev)
    
    def close(self):
        """
        Close the search structure to release allocated memory.
        This method shall be called when done with text searching.
        """
        pdfium.FPDFText_FindClose(self._search)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import os
import os.path
import ctypes
import logging
import functools
from concurrent.futures import ProcessPoolExecutor

import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers._utils import ViewmodeMapping
from pypdfium2._helpers._opener import (
    open_pdf,
    is_input_buffer,
)
from pypdfium2._helpers.page import PdfPage
from pypdfium2._helpers.misc import (
    OutlineItem,
    FileAccess,
    PdfiumError,
)

try:
    import uharfbuzz as harfbuzz
except ImportError:
    have_harfbuzz = False
else:
    have_harfbuzz = True

logger = logging.getLogger(__name__)


class PdfDocument:
    """
    Document helper class.
    
    Parameters:
        input_data (str | bytes | typing.BinaryIO | FPDF_DOCUMENT):
            The input PDF given as file path, bytes, byte buffer, or raw PDFium document handle.
            A byte buffer is defined as an object that implements the methods ``seek()``, ``tell()`` and ``readinto()``.
        password (str | bytes):
            A password to unlock the PDF, if encrypted.
        file_access (FileAccess):
            Define how files shall be opened internally. This parameter is ignored if *input_data* is not a file path.
        autoclose (bool):
            If set to :data:`True` and a byte buffer was provided as input, :meth:`.close` will not only close the PDFium document, but also the input source.
    
    Hint:
        * :func:`len` may be called to get a document's number of pages.
        * :class:`PdfDocument` implements the context manager API, hence documents can be used in a ``with`` block, where :meth:`.close` will be called automatically on exit.
        * Looping over a document will yield its pages from beginning to end.
        * Pages may be loaded using list index access.
        * The ``del`` keyword and list index access may be used to delete pages.
    
    Note:
        :class:`.PdfDocument` does not implement a page cache. This is to ensure correct behaviour in case the order or number of pages is modified using the raw API.
        Therefore, it is up to the caller to cache pages appropriately and avoid inefficient repated loading/closing of pages.
    """
    
    def __init__(
            self,
            input_data,
            password = None,
            file_access = FileAccess.NATIVE,
            autoclose = False,
        ):
        
        self._orig_input = input_data
        self._actual_input = input_data
        self._rendering_input = None
        self._ld_data = None
        
        self._password = password
        self._file_access = file_access
        self._autoclose = autoclose
        
        if isinstance(self._orig_input, str):
            
            self._orig_input = os.path.abspath( os.path.expanduser(self._orig_input) )
            if not os.path.isfile(self._orig_input):
                raise FileNotFoundError("File does not exist: '%s'" % self._orig_input)
            
            if self._file_access is FileAccess.NATIVE:
                pass
            elif self._file_access is FileAccess.BUFFER:
                self._actual_input = open(self._orig_input, "rb")
                self._autoclose = True
            elif self._file_access is FileAccess.BYTES:
                buf = open(self._orig_input, "rb")
                self._actual_input = buf.read()
                buf.close()
            else:
                raise ValueError("An invalid file access strategy was given: '%s'" % self._file_access)
        
        if isinstance(self._actual_input, pdfium.FPDF_DOCUMENT):
            self._pdf = self._actual_input
        else:
            self._pdf, self._ld_data = open_pdf(self._actual_input, self._password, self._autoclose)
    
    
    def __enter__(self):
        return self
    
    def __exit__(self, *_):
        self.close()
    
    def __len__(self):
        return pdfium.FPDF_GetPageCount(self._pdf)
    
    def __iter__(self):
        for i in range( len(self) ):
            yield self.get_page(i)
    
    def __getitem__(self, i):
        return self.get_page(i)
    
    def __delitem__(self, i):
        self.del_page(i)
    
    
    @property
    def raw(self):
        """ FPDF_DOCUMENT: The raw PDFium document object handle. """
        return self._pdf
    
    @classmethod
    def new(cls):
        """
        Returns:
            PdfDocument: A new, empty document.
        """
        new_pdf = pdfium.FPDF_CreateNewDocument()
        return cls(new_pdf)
    
    def close(self):
        """
        Close the document to release allocated memory.
        This function shall be called when finished working with the object.
        """
        pdfium.FPDF_CloseDocument(self._pdf)
        if self._ld_data is not None:
            self._ld_data.close()
    
    
    def save(self, buffer, version=None):
        """
        Save the document into an output buffer, at its current state.
        
        Parameters:
            buffer (typing.BinaryIO):
                A byte buffer to capture the data.
                It may be any object implementing the ``write()`` method.
            version (typing.Optional[int]):
                 The PDF version to use, given as an integer (14 for 1.4, 15 for 1.5, ...).
                 If :data:`None`, PDFium will set a version automatically.
        """
        
        WriteFunctype = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(pdfium.FPDF_FILEWRITE), ctypes.POINTER(None), ctypes.c_ulong)
        
        filewrite = pdfium.FPDF_FILEWRITE()
        filewrite.WriteBlock = WriteFunctype( _writer_class(buffer) )
        
        saveargs = (self._pdf, ctypes.byref(filewrite), pdfium.FPDF_NO_INCREMENTAL)
        if version is None:
            success = pdfium.FPDF_SaveAsCopy(*saveargs)
        else:
            success = pdfium.FPDF_SaveWithVersion(*saveargs, version)
        
        if not success:
            raise PdfiumError("Saving the document failed")
    
    
    def _verify_index(self, index):
        n_pages = len(self)
        if not 0 <= index < n_pages:
            raise IndexError("Page index %s is out of bounds for document with %s pages." % (index, n_pages))
    
    def new_page(self, width, height, index=None):
        """
        Insert a new, empty page into the document.
        
        Parameters:
            width (float):
                Target page width (horizontal size).
            height (float):
                Target page height (vertical size).
            index (typing.Optional[int]):
                Suggested zero-based index at which the page will be inserted.
                If *index* is less or equal to zero, the page will be inserted at the beginning.
                If *index* is :data:`None` or larger that the document's current last index, the page will be appended to the end.
        
        Returns:
            PdfPage: The newly created page.
        """
        if index is None:
            index = len(self)
        raw_page = pdfium.FPDFPage_New(self._pdf, index, width, height)
        return PdfPage(raw_page, self)
    
    def del_page(self, index):
        """ Remove the page at *index*. """
        self._verify_index(index)
        pdfium.FPDFPage_Delete(self._pdf, index)
    
    def get_page(self, index):
        """
        Returns:
            PdfPage: The page at *index*.
        """
        self._verify_index(index)
        raw_page = pdfium.FPDF_LoadPage(self._pdf, index)
        return PdfPage(raw_page, self)
    
    
    def add_font(self, font_path, type, is_cid):
        """
        Add a font to the document.
        
        Parameters:
            font_path (str):
                File path of the font to use.
            type (int):
                A constant signifying the type of the given font (:data:`.FPDF_FONT_TYPE1` or :data:`.FPDF_FONT_TRUETYPE`).
            is_cid (bool):
                Whether the given font is a CID font or not.
        Returns:
            PdfFont: A PDF font helper object.
        """
        
        with open(font_path, "rb") as fh:
            font_data = fh.read()
        
        pdf_font = pdfium.FPDFText_LoadFont(
            self._pdf,
            ctypes.cast(font_data, ctypes.POINTER(ctypes.c_uint8)),
            len(font_data),
            type,
            is_cid,
        )
        
        return PdfFont(pdf_font, font_data)
    
    
    def _get_bookmark(self, bookmark, level):
        
        t_buflen = pdfium.FPDFBookmark_GetTitle(bookmark, None, 0)
        t_buffer = ctypes.create_string_buffer(t_buflen)
        pdfium.FPDFBookmark_GetTitle(bookmark, t_buffer, t_buflen)
        title = t_buffer.raw[:t_buflen].decode('utf-16-le')[:-1]
        
        is_closed = pdfium.FPDFBookmark_GetCount(bookmark) < 0
        dest = pdfium.FPDFBookmark_GetDest(self._pdf, bookmark)
        page_index = pdfium.FPDFDest_GetDestPageIndex(self._pdf, dest)
        
        n_params = ctypes.c_ulong()
        view_pos = (pdfium.FS_FLOAT * 4)()
        view_mode = pdfium.FPDFDest_GetView(dest, n_params, view_pos)
        view_pos = list(view_pos)[:n_params.value]
        
        return OutlineItem(
            level = level,
            title = title,
            is_closed = is_closed,
            page_index = page_index,
            view_mode = view_mode,
            view_pos = view_pos,
        )
    
    
    def get_toc(
            self,
            max_depth = 15,
            parent = None,
            level = 0,
            seen = None,
        ):
        """
        Read the document's outline ("table of contents").
        
        Parameters:
            max_depth (int):
                Maximum recursion depth to consider when reading the outline.
        Yields:
            :class:`.OutlineItem`: The data of an outline item ("bookmark").
        """
        
        if level >= max_depth:
            return []
        if seen is None:
            seen = set()
        
        bookmark = pdfium.FPDFBookmark_GetFirstChild(self._pdf, parent)
        
        while bookmark:
            
            address = ctypes.addressof(bookmark.contents)
            if address in seen:
                logger.warning("A circular bookmark reference was detected whilst parsing the table of contents.")
                break
            else:
                seen.add(address)
            
            yield self._get_bookmark(bookmark, level)
            yield from self.get_toc(
                max_depth = max_depth,
                parent = bookmark,
                level = level + 1,
                seen = seen,
            )
            
            bookmark = pdfium.FPDFBookmark_GetNextSibling(self._pdf, bookmark)
    
    
    @staticmethod
    def print_toc(toc, n_digits=2):
        """
        Print a table of contents.
        
        Parameters:
            toc (typing.Iterator[OutlineItem]):
                Sequence of outline items to show.
            n_digits (int):
                The number of digits to which viewport coordinates shall be rounded.
        """
        
        for item in toc:
            print(
                "    " * item.level +
                "[%s] " % ("-" if item.is_closed else "+") +
                "%s -> %s  # %s %s" % (
                    item.title,
                    item.page_index+1,
                    ViewmodeMapping[item.view_mode],
                    [round(c, n_digits) for c in item.view_pos],
                )
            )
    
    
    def update_rendering_input(self):
        """
        Update the input sources for concurrent rendering to the document's current state
        by saving to bytes and setting the result as new input.
        If you modified the document, you may want to call this method before :meth:`._render_base`.
        """
        buffer = io.BytesIO()
        self.save(buffer)
        buffer.seek(0)
        self._rendering_input = buffer.read()
        buffer.close()
    
    
    @classmethod
    def _process_page(cls, index, renderer_name, input_data, password, file_access, **kwargs):
        pdf = cls(
            input_data,
            password = password,
            file_access = file_access,
        )
        page = pdf.get_page(index)
        result = getattr(page, "render_to"+renderer_name)(**kwargs)
        [g.close() for g in (page, pdf)]
        return result, index
    
    
    def _render_base(
            self,
            renderer_name,
            page_indices = None,
            n_processes = os.cpu_count(),
            **kwargs
        ):
        """
        Concurrently render multiple pages, using a process pool executor.
        This method serves as base for :meth:`.render_tobytes` and :meth:`render_topil`. Embedders should never call it directly.
        
        The order of results matches the order of given page indices.
        
        Parameters:
            page_indices (typing.Optional[typing.Sequence[int]]):
                A sequence of zero-based indices of the pages to render.
                If :data:`None`, all pages will be included.
            n_processes (int):
                Target number of parallel processes.
            kwargs (dict):
                Keyword arguments to be passed to :meth:`.PdfPage.render_base`.
        Yields:
            :data:`typing.Any`: Implementation-specific result object.
        """
        
        if self._rendering_input is None:
            if isinstance(self._orig_input, pdfium.FPDF_DOCUMENT):
                logger.warning("Cannot perform concurrent processing without input sources - saving the document implicitly to get picklable data.")
                self.update_rendering_input()
            elif is_input_buffer(self._orig_input):
                logger.warning("Cannot perform concurrent rendering with buffer input - reading the whole buffer into memory implicitly.")
                self._orig_input.seek(0)
                self._rendering_input = self._orig_input.read()
            else:
                self._rendering_input = self._orig_input
        
        n_pages = len(self)
        
        if page_indices is None or len(page_indices) == 0:
            page_indices = [i for i in range(n_pages)]
        if not all(0 <= i < n_pages for i in page_indices):
            raise ValueError("A page index is out of bounds.")
        
        invoke_renderer = functools.partial(
            PdfDocument._process_page,
            renderer_name = renderer_name,
            input_data = self._rendering_input,
            password = self._password,
            file_access = self._file_access,
            **kwargs
        )
        
        i = 0
        with ProcessPoolExecutor(n_processes) as pool:
            for result, index in pool.map(invoke_renderer, page_indices):
                assert index == page_indices[i]
                i += 1
                yield result
        
        assert len(page_indices) == i
    
    
    def render_tobytes(self, **kwargs):
        """
        Concurrently render pages to bytes.
        See :meth:`.PdfDocument._render_base` and :meth:`.PdfPage.render_base` for possible keyword arguments.
        
        Yields:
            :class:`tuple`: Result of :meth:`.PdfPage.render_tobytes`.
        """
        yield from self._render_base("bytes", **kwargs)
    
    def render_topil(self, **kwargs):
        """
        *Requires* :mod:`PIL`.
        
        Concurrently render pages to PIL images.
        See :meth:`.PdfDocument._render_base` and :meth:`.PdfPage.render_base` for possible keyword arguments.
        
        Yields:
            :class:`PIL.Image.Image`: PIL image.
        """
        yield from self._render_base("pil", **kwargs)


class _writer_class:
    
    def __init__(self, buffer):
        self.buffer = buffer
        if not callable( getattr(self.buffer, "write", None) ):
            raise ValueError("Output buffer must implement the write() method.")
    
    def __call__(self, _, data, size):
        block = ctypes.cast(data, ctypes.POINTER(ctypes.c_ubyte * size))
        self.buffer.write(block.contents)
        return 1


class HarfbuzzFont:
    """ Harfbuzz font data helper class. """
    
    def __init__(self, font_path):
        if not have_harfbuzz:
            raise RuntimeError("Font helpers require uharfbuzz to be installed.")
        self.blob = harfbuzz.Blob.from_file_path(font_path)
        self.face = harfbuzz.Face(self.blob)
        self.font = harfbuzz.Font(self.face)
        self.scale = self.font.scale[0]


class PdfFont:
    """ PDF font data helper class. """
    
    def __init__(self, pdf_font, font_data):
        self._pdf_font = pdf_font
        self._font_data = font_data
    
    @property
    def raw(self):
        """ FPDF_FONT: The raw PDFium font object handle. """
        return self._pdf_font
    
    def close(self):
        """
        Close the font to release allocated memory.
        This function shall be called when finished working with the object.
        """
        pdfium.FPDFFont_Close(self._pdf_font)
        id(self._font_data)



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import shutil
from os.path import dirname, abspath, join

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import SourceTree, run_cmd


RepositoryURL = "https://github.com/pypdfium2-team/pypdfium2"

def get_url(tag):
    return RepositoryURL + "/tree/%s" % tag

def generate_tags_list(Git, count=10):
    tags_str = run_cmd([Git, "for-each-ref", "--count=%s" % count, "--sort=-creatordate", "--format", "%(refname)", "refs/tags"], cwd=SourceTree, capture=True)
    tags_list = tags_str.split("\n")
    return tags_list


def get_tag(tags_list, n_descends, skip_beta=False):
    i = 0
    for line in tags_list:
        tag = line.split("/")[-1].strip()
        if skip_beta and i > 0 and "b" in tag:
            continue
        if i >= n_descends:
            break
        i += 1
    return tag


def main():
    
    Git = shutil.which("git")
    tags_list = generate_tags_list(Git)
    current_tag, prev_tag = get_tag(tags_list, 0), get_tag(tags_list, 1, True)
    print(current_tag, prev_tag)
    
    relnotes = "Release %s\n\n" % current_tag
    relnotes += "### Changes\n\nCommits between [`%s`](%s) and [`%s`](%s):\n\n" % (prev_tag, get_url(prev_tag), current_tag, get_url(current_tag))
    relnotes += run_cmd([Git, "log", "%s..%s" % (prev_tag, current_tag), "--pretty=format:* %H %s"], cwd=SourceTree, capture=True)
    relnotes += "\n"
    
    with open(join(SourceTree, "RELEASE.md"), "w") as fh:
        fh.write(relnotes)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import re
import pytest
import sysconfig
import configparser
from os.path import join
from pathlib import Path
from wheel.bdist_wheel import bdist_wheel
from pl_setup import (
    setup_base,
    packaging_base as pkg_base,
)
from pl_setup.packaging_base import PlatformNames
from pl_setup import update_pdfium as fpdf_up
from .conftest import pl_names, SourceTree


# module

def test_entrypoint():
    
    setup_cfg = configparser.ConfigParser()
    setup_cfg.read( join(SourceTree,'setup.cfg') )
    console_scripts = setup_cfg['options.entry_points']['console_scripts']
    
    entry_point = console_scripts.split('=')[-1].strip().split(':')
    module_path = entry_point[0]
    method_name = entry_point[1]
    
    namespace = {}
    exec("from %s import %s" % (module_path, method_name), namespace)
    assert method_name in namespace
    
    function = namespace[method_name]
    assert callable(function)


# setup_base

ExpectedTags = (
    (PlatformNames.darwin_x64,    "macosx_10_11_x86_64.macosx_11_0_x86_64.macosx_12_0_x86_64"),
    (PlatformNames.darwin_arm64,  "macosx_11_0_arm64.macosx_12_0_arm64"),
    (PlatformNames.linux_x64,     "manylinux_2_17_x86_64.manylinux2014_x86_64"),
    (PlatformNames.linux_x86,     "manylinux_2_17_i686.manylinux2014_i686"),
    (PlatformNames.linux_arm64,   "manylinux_2_17_aarch64.manylinux2014_aarch64"),
    (PlatformNames.linux_arm32,   "manylinux_2_17_armv7l.manylinux2014_armv7l"),
    (PlatformNames.musllinux_x64, "musllinux_1_2_x86_64"),
    (PlatformNames.musllinux_x86, "musllinux_1_2_i686"),
    (PlatformNames.windows_x64,   "win_amd64"),
    (PlatformNames.windows_arm64, "win_arm64"),
    (PlatformNames.windows_x86,   "win32"),
    (PlatformNames.sourcebuild,   sysconfig.get_platform().replace('-','_').replace('.','_')),
)


def test_expected_tags():
    assert len(pl_names) == len(ExpectedTags)
    for platform, tag in ExpectedTags:
        assert hasattr(PlatformNames, platform)
        assert isinstance(tag, str)


def test_get_tag():
    for platform, tag in ExpectedTags:
        assert setup_base._get_tag(platform) == tag

def test_unknown_tag():
    plat_dir = "win_amd74"
    with pytest.raises(ValueError, match=re.escape("Unknown platform directory %s" % plat_dir)):
        setup_base._get_tag(plat_dir)

def test_get_bdist():
    for platform, tag in ExpectedTags:
        bdist_cls = setup_base._get_bdist(platform)
        assert issubclass(bdist_cls, bdist_wheel)


# packaging_base

def test_libnames():
    for name in pkg_base.Libnames:
        assert "pdfium" in name

def test_platformnames():
    for member in pl_names:
        assert member == getattr(pkg_base.PlatformNames, member)

def test_paths():
    assert pkg_base.HomeDir == str( Path.home() )
    assert pkg_base.SourceTree == SourceTree
    assert pkg_base.DataTree == str( Path(SourceTree) / "data" )
    assert pkg_base.SB_Dir == str( Path(SourceTree) / "sourcebuild" )
    assert pkg_base.ModuleDir == str( Path(SourceTree) / "src" / "pypdfium2" )
    assert pkg_base.VersionFile == str( Path(pkg_base.ModuleDir) / "version.py" )


# update_pdfium

def test_releasenames():
    assert len(fpdf_up.ReleaseNames) == len(pl_names) - 1
    for key, value in fpdf_up.ReleaseNames.items():
        assert hasattr(PlatformNames, key)
        prefix, system, cpu = value.replace("linux-musl", "musllinux").split("-", maxsplit=3)
        assert prefix == "pdfium"
        assert system in ("linux", "musllinux", "mac", "win")
        assert cpu in ("x64", "x86", "arm64", "arm")



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pytest
import pypdfium2 as pdfium
from ..conftest import TestFiles


def test_boxes():
    
    pdf = pdfium.PdfDocument(TestFiles.render)
    page = pdf.get_page(0)
    assert page.get_size() == (595, 842)
    assert page.get_mediabox() == (0, 0, 595, 842)
    assert isinstance(page, pdfium.PdfPage)
    
    test_cases = [
        ("media", (0,  0,  612, 792)),
        ("media", (0,  0,  595, 842)),
        ("crop",  (10, 10, 585, 832)),
        ("bleed", (20, 20, 575, 822)),
        ("trim",  (30, 30, 565, 812)),
        ("art",   (40, 40, 555, 802)),
    ]
    
    for meth_name, exp_box in test_cases:
        getattr(page, "set_%sbox" % meth_name)(*exp_box)
        box = getattr(page, "get_%sbox" % meth_name)()
        assert pytest.approx(box) == exp_box
    
    [g.close() for g in (page, pdf)]


def test_mediabox_fallback():
    pdf = pdfium.PdfDocument(TestFiles.box_fallback)
    page = pdf.get_page(0)
    assert page.get_mediabox() == (0, 0, 612, 792)
    [g.close() for g in (page, pdf)]


def test_rotation():
    pdf = pdfium.PdfDocument.new()
    page = pdf.new_page(500, 800)
    for r in (90, 180, 270, 0):
        page.set_rotation(r)
        assert page.get_rotation() == r
    [g.close() for g in (page, pdf)]


def test_pageobjects():
    pdf = pdfium.PdfDocument(TestFiles.images)
    page = pdf.get_page(0)
    
    images = [obj for obj in page.get_objects() if obj.get_type() == pdfium.FPDF_PAGEOBJ_IMAGE]
    assert len(images) == 3
    
    positions = [img.get_pos() for img in images]
    exp_positions = [
        (133, 459, 350, 550),
        (48, 652, 163, 700),
        (204, 204, 577, 360),
    ]
    assert len(positions) == len(exp_positions)
    for pos, exp_pos in zip(positions, exp_positions):
        assert pytest.approx(pos, abs=1) == exp_pos
    
    [g.close() for g in (page, pdf)]



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2 import _namespace as pdfium


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "toc",
        help = "Show a PDF document's table of contents",
    )
    parser.add_argument(
        "input",
        help = "PDF document of which to print the outline",
    )
    parser.add_argument(
        "--max-depth",
        type = int,
        default = 15,
        help = "Maximum recursion depth to consider when parsing the table of contents",
    )


def main(args):
    pdf = pdfium.PdfDocument(args.input)
    pdf.print_toc( pdf.get_toc() )
    pdf.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import math
import logging
import PIL.Image
from os.path import join
import pytest
import pypdfium2 as pdfium
from ..conftest import TestFiles, OutputDir, ExpRenderPixels, get_members


@pytest.fixture
def sample_page():
    pdf = pdfium.PdfDocument(TestFiles.render)
    page = pdf.get_page(0)
    yield page
    [g.close() for g in (page, pdf)]


@pytest.fixture
def multipage_doc():
    pdf = pdfium.PdfDocument(TestFiles.multipage)
    yield pdf
    pdf.close()


def _check_pixels(pil_image, pixels):
    for pos, value in pixels:
        assert pil_image.getpixel(pos) == value


@pytest.mark.parametrize(
    ("name", "crop", "scale", "rotation"),
    [   
        ["01_r0",      (0,   0,   0,   0  ), 0.25, 0,   ],
        ["02_r90",     (0,   0,   0,   0  ), 0.5,  90,  ],
        ["03_r180",    (0,   0,   0,   0  ), 0.75, 180, ],
        ["04_r270",    (0,   0,   0,   0  ), 1,    270, ],
        ["05_cl",      (100, 0,   0,   0  ), 0.5,  0,   ],
        ["06_cb",      (0,   100, 0,   0  ), 0.5,  0,   ],
        ["07_cr",      (0,   0,   100, 0  ), 0.5,  0,   ],
        ["08_ct",      (0,   0,   0,   100), 0.5,  0,   ],
        ["09_r90_cb",  (0,   100,  0,  0  ), 0.5,  90,  ],
        ["10_r180_cr", (0,   0,   100, 0  ), 0.5,  180, ],
        ["11_r270_ct", (0,   0,   0,   100), 0.5,  270, ],
    ]
)
def test_render_page_transform(sample_page, name, crop, scale, rotation):
    pil_image = sample_page.render_topil(
        crop = crop,
        scale = scale,
        rotation = rotation,
    )
    pil_image.save( join(OutputDir, "%s.png" % name) )
    assert pil_image.mode == "RGB"
    
    c_left, c_bottom, c_right, c_top = [math.ceil(c*scale) for c in crop]
    w = math.ceil(595*scale)
    h = math.ceil(842*scale)
    if rotation in (90, 270):
        w, h = h, w
    
    c_w = w - c_left - c_right
    c_h = h - c_bottom - c_top
    assert pil_image.size == (c_w, c_h)
    
    pixels = []
    for (x, y), value in ExpRenderPixels:
        x, y = round(x*scale), round(y*scale)
        if rotation in (90, 270):
            x, y = y, x
        if rotation == 90:
            x = w-1 - x
        elif rotation == 180:
            x = w-1 - x
            y = h-1 - y
        elif rotation == 270:
            y = h-1 - y
        x -= c_left
        y -= c_top
        if 0 <= x < c_w and 0 <= y < c_h:
            pixels.append( ((x, y), value) )
    
    _check_pixels(pil_image, pixels)
    pil_image.close()


def test_render_page_alpha(sample_page):
    
    pixels = [
        [(0,   0  ), (0,   0,   0,   0  )],
        [(62,  66 ), (0,   0,   0,   186)],
        [(150, 180), (129, 212, 26,  255)],
        [(150, 390), (42,  96,  153, 255)],
        [(150, 570), (128, 0,   128, 255)],
    ]
    pil_image = sample_page.render_topil(colour=None)
    
    assert pil_image.mode == "RGBA"
    assert pil_image.size == (595, 842)
    for pos, exp_value in pixels:
        assert pil_image.getpixel(pos) == exp_value
    
    pil_image.close()


def test_render_page_greyscale(sample_page):

    image_a = sample_page.render_topil(
        greyscale = True,
        scale = 0.5
    )
    assert image_a.size == (298, 421)
    assert image_a.mode == "L"
    image_a.save(join(OutputDir, "greyscale.png"))
    image_a.close()
    
    # PDFium currently doesn't support LA, hence greyscale + alpha rendering results in RGBA
    image_b = sample_page.render_topil(
        greyscale = True,
        colour = None,
        scale = 0.5
    )
    assert image_b.size == (298, 421)
    assert image_b.mode == "RGBA"
    image_b.save(join(OutputDir, "greyscale_alpha.png"))
    image_b.close()


@pytest.mark.parametrize(
    "colour",
    [
        (255, 255, 255, 255),
        (60,  70,  80,  100),
        (255, 255, 255),
        (0,   255, 255),
        (255, 0,   255),
        (255, 255, 0  ),
    ]
)
def test_render_page_bgcolour(colour, sample_page):
    
    pil_image = sample_page.render_topil(colour=colour, scale=0.5)
    assert pil_image.size == (298, 421)
    
    px_colour = colour
    if len(colour) == 4:
        if colour[3] == 255:
            px_colour = colour[:-1]
    
    bg_pixel = pil_image.getpixel( (0, 0) )
    assert bg_pixel == px_colour
    
    pil_image.close()


def test_render_page_tobytes(sample_page):
    
    bytedata, cl_format, size = sample_page.render_tobytes(scale=0.5)
    
    assert isinstance(bytedata, bytes)
    assert cl_format == "BGR"
    assert size == (298, 421)
    assert len(bytedata) == size[0] * size[1] * len(cl_format)
    
    pil_image = PIL.Image.frombytes("RGB", size, bytedata, "raw", "BGR")
    assert pil_image.mode == "RGB"
    assert pil_image.size == (298, 421)
    assert isinstance(pil_image, PIL.Image.Image)
    pil_image.close()


def test_render_page_optimisation(sample_page):
    
    modes = get_members(pdfium.OptimiseMode)
    assert len(modes) == 3
    assert set(modes) == set([
        pdfium.OptimiseMode.NONE,
        pdfium.OptimiseMode.LCD_DISPLAY,
        pdfium.OptimiseMode.PRINTING,
    ])
    
    for mode in modes:
        pil_image = sample_page.render_topil(
            optimise_mode = mode,
            scale = 0.5,
        )
        assert isinstance(pil_image, PIL.Image.Image)
        pil_image.close()


def test_render_page_noantialias(sample_page):
    pil_image = sample_page.render_topil(
        no_antialias = ("text", "image", "path"),
        scale = 0.5,
    )
    assert isinstance(pil_image, PIL.Image.Image)
    pil_image.close()


@pytest.fixture
def render_pdffile_topil(multipage_doc):
    
    renderer = multipage_doc.render_topil(
        scale=0.5, greyscale=True,
    )
    imgs = []
    
    for image in renderer:
        assert isinstance(image, PIL.Image.Image)
        assert image.mode == "L"
        imgs.append(image)
    
    assert len(imgs) == 3
    yield imgs
    [image.close() for image in imgs]


@pytest.fixture
def render_pdffile_tobytes(multipage_doc):
    
    renderer = multipage_doc.render_tobytes(
        scale=0.5, greyscale=True,
    )
    imgs = []
    
    for imgdata, cl_format, size in renderer:
        assert cl_format == "L"
        assert isinstance(imgdata, bytes)
        assert len(imgdata) == size[0] * size[1] * len(cl_format)
        pil_image = PIL.Image.frombytes("L", size, imgdata, "raw", "L")
        imgs.append(pil_image)
    
    assert len(imgs) == 3
    yield imgs
    [image.close() for image in imgs]


def test_render_pdffile(render_pdffile_topil, render_pdffile_tobytes):
    for image_a, image_b in zip(render_pdffile_topil, render_pdffile_tobytes):
        assert image_a == image_b


def test_render_pdf_new(caplog):
    
    pdf = pdfium.PdfDocument.new()
    page = pdf.new_page(50, 100)
    
    with caplog.at_level(logging.WARNING):
        renderer = pdf.render_topil()
        image = next(renderer)
    
    warning = "Cannot perform concurrent processing without input sources - saving the document implicitly to get picklable data."
    assert warning in caplog.text
    
    assert isinstance(image, PIL.Image.Image)
    assert image.mode == "RGB"
    assert image.size == (50, 100)
    
    [g.close() for g in (image, page, pdf)]


def test_render_pdfbuffer(caplog):
    
    buffer = open(TestFiles.render, "rb")
    pdf = pdfium.PdfDocument(buffer)
    assert pdf._orig_input is buffer
    assert pdf._actual_input is buffer
    assert pdf._rendering_input is None
    
    with caplog.at_level(logging.WARNING):
        for image in pdf.render_topil(scale=0.5):
            assert isinstance(image, PIL.Image.Image)
    
    assert isinstance(pdf._rendering_input, bytes)
    warning = "Cannot perform concurrent rendering with buffer input - reading the whole buffer into memory implicitly."
    assert warning in caplog.text
    
    pdf.close()
    buffer.close()


def test_render_pdfbytes():
    
    with open(TestFiles.render, "rb") as fh:
        data = fh.read()
    
    pdf = pdfium.PdfDocument(data)
    assert pdf._orig_input is data
    assert pdf._actual_input is data
    assert pdf._rendering_input is None
    for image in pdf.render_topil(scale=0.5):
        assert isinstance(image, PIL.Image.Image)
    assert isinstance(pdf._rendering_input, bytes)
    
    pdf.close()


def test_render_pdffile_asbuffer():
    
    pdf = pdfium.PdfDocument(TestFiles.render, file_access=pdfium.FileAccess.BUFFER)
    
    assert pdf._orig_input == TestFiles.render
    assert isinstance(pdf._actual_input, io.BufferedReader)
    assert pdf._rendering_input is None
    assert pdf._file_access is pdfium.FileAccess.BUFFER
    
    for image in pdf.render_topil(scale=0.5):
        assert isinstance(image, PIL.Image.Image)
    
    # Not sure how to test that the requested file access strategy is actually used when constructing the new PdfDocument objects
    assert pdf._rendering_input == TestFiles.render
    
    pdf.close()
    assert pdf._actual_input.closed is True


def test_render_pdffile_asbytes():
    
    pdf = pdfium.PdfDocument(TestFiles.render, file_access=pdfium.FileAccess.BYTES)
    
    assert pdf._orig_input == TestFiles.render
    assert isinstance(pdf._actual_input, bytes)
    assert pdf._rendering_input is None
    assert pdf._file_access is pdfium.FileAccess.BYTES
    
    for image in pdf.render_topil(scale=0.5):
        assert isinstance(image, PIL.Image.Image)
    assert pdf._rendering_input == TestFiles.render
    
    pdf.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os.path
from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parsers import pagetext_type
from pypdfium2._helpers._utils import (
    ObjtypeToConst,
    ObjtypeToName,
)


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "find-pageobjects",
        help = "Locate page objects of a certain type",
    )
    parser.add_argument(
        "input",
        type = os.path.abspath,
        help = "Path to the PDF document to work with",
    )
    parser.add_argument(
        "--password",
        help = "Password to unlock the PDF, if encrypted"
    )
    parser.add_argument(
        "--pages",
        type = pagetext_type,
        help = "The pages to search (defaults to all)",
    )
    parser.add_argument(
        "--types",
        nargs = "+",
        required = True,
        choices = [k for k in ObjtypeToConst.keys()],
        help = "Object types to consider",
    )


def main(args):
    
    doc = pdfium.PdfDocument(args.input, args.password)
    args.types = [ObjtypeToConst[t] for t in args.types]
    if args.pages is None:
        args.pages = [i for i in range(len(doc))]
    
    for index in args.pages:
        page = doc.get_page(index)
        for obj in page.get_objects():
            type = obj.get_type()
            if type in args.types:
                print(ObjtypeToName[type], obj.get_pos())
        page.close()
    
    doc.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# Non-stdlib imports not allowed in this file, as it is imported prior to the check_deps call

import shutil
import subprocess
from glob import glob
from os.path import (
    expanduser,
    dirname,
    abspath,
    join,
)

Libnames = (
    "pdfium",
    "pdfium.dylib",
    "pdfium.dll",
    "libpdfium.so",
    "pdfium.so",
    "libpdfium",
    "libpdfium.dylib",
    "libpdfium.dll",
)


HomeDir     = expanduser("~")
SourceTree  = dirname(dirname(dirname(abspath(__file__))))
DataTree    = join(SourceTree, "data")
SB_Dir      = join(SourceTree, "sourcebuild")
ModuleDir   = join(SourceTree, "src", "pypdfium2")
VersionFile = join(ModuleDir, "version.py")


class PlatformNames:
    darwin_x64    = "darwin_x64"
    darwin_arm64  = "darwin_arm64"
    linux_x64     = "linux_x64"
    linux_x86     = "linux_x86"
    linux_arm64   = "linux_arm64"
    linux_arm32   = "linux_arm32"
    musllinux_x64 = "musllinux_x64"
    musllinux_x86 = "musllinux_x86"
    windows_x64   = "windows_x64"
    windows_x86   = "windows_x86"
    windows_arm64 = "windows_arm64"
    sourcebuild   = "sourcebuild"


def run_cmd(command, cwd, capture=False, **kwargs):
    
    print('%s ("%s")' % (command, cwd))
    if capture:
        kwargs.update( dict(stdout=subprocess.PIPE, stderr=subprocess.STDOUT) )
    
    comp_process = subprocess.run(command, cwd=cwd, **kwargs)
    if capture:
        return comp_process.stdout.decode("UTF-8").strip()
    else:
        return comp_process


def call_ctypesgen(platform_dir, include_dir):
    
    bindings_file = join(platform_dir, "_pypdfium.py")
    
    ctypesgen_cmd = [shutil.which("ctypesgen"), "--library", "pdfium", "--strip-build-path", platform_dir, "-L", "."] + sorted(glob( join(include_dir, "*.h") )) + ["-o", bindings_file]
    run_cmd(ctypesgen_cmd, cwd=platform_dir)
    
    with open(bindings_file, "r") as file_reader:
        text = file_reader.read()
        text = text.replace(platform_dir, ".")
        text = text.replace(HomeDir, "~")
    
    with open(bindings_file, "w") as file_writer:
        file_writer.write(text)


def get_version_ns():
    ver_ns = {}
    with open(VersionFile, "r") as fh:
        exec(fh.read(), ver_ns)
    ver_ns = {k: v for k, v in ver_ns.items() if not k.startswith("_")}
    return ver_ns

VerNamespace = get_version_ns()


def set_version(variable, new_ver):
    
    with open(VersionFile, "r") as fh:
        content = fh.read()
    
    if isinstance(new_ver, str):
        template = '%s = "%s"'
    else:
        template = '%s = %s'
    previous = template % (variable, VerNamespace[variable])
    updated = template % (variable, new_ver)
    
    print("'%s' -> '%s'" % (previous, updated))
    assert content.count(previous) == 1
    content = content.replace(previous, updated)
    VerNamespace[variable] = new_ver
    
    with open(VersionFile, "w") as fh:
        fh.write(content)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
from pypdfium2 import _namespace as pdfium


def _merge_files(input_paths):
    
    dest_pdf = pdfium.PdfDocument.new()
    
    for in_path in reversed(input_paths):
        src_pdf = pdfium.PdfDocument(in_path)
        n_pages = len(src_pdf)
        page_indices = (ctypes.c_int * n_pages)(*[i for i in range(n_pages)])
        pdfium.FPDF_ImportPagesByIndex(dest_pdf.raw, src_pdf.raw, page_indices, n_pages, 0)
        src_pdf.close()
    
    return dest_pdf


def attach_parser(subparsers):
    parser = subparsers.add_parser(
        "merge",
        help = "Concatenate PDF files",
    )
    parser.add_argument(
        "inputs",
        nargs = "+",
        help = "A sequence of PDF files to concatenate",
    )
    parser.add_argument(
        "--output", "-o",
        required = True,
        help = "Target path for the output document",
    )


def main(args):
    merged_pdf = _merge_files(args.inputs)
    with open(args.output, 'wb') as buffer:
        merged_pdf.save(buffer)
    merged_pdf.close()
